;################################################################################
;#										#
;# chipbasic - single chip basic computer with ATMega 16			#
;# line compressor/expander							#
;# copyright (c) 2007 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;#										#
;# This program is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU General Public License			#
;# as published by the Free Software Foundation; either version 2		#
;# of the License, or (at your option) any later version.			#
;#										#
;# This program is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################
.set	lce_sourcelen = 27			;max length of text buffer
.set	lce_error = 13				;error: line too long

;-----------------------------------------------------------------------
; compress inbuf->linebuf
;-----------------------------------------------------------------------
lce_comp:	ldi	r17,0			;destcounter
		ldi	r18,0			;normal mode
		ldi	YL,LOW(bas_inbuf+3)	;text buffer address
		ldi	YH,HIGH(bas_inbuf+3)
		ldi	XL,LOW(bas_linebuf)	;code buffer address
		ldi	XH,HIGH(bas_linebuf)
		ldi	r19,':'			;this is only virtual 
		
lce_comp_mloop:	mov	tempreg5,YL		;save source pointer
		mov	tempreg6,YH

lce_comp_loop:	cpi	r17,mem_linelen		;end of code
		brne	lce_comp_01		;no
		ldi	ereg,lce_error		;line too long		
		ret

lce_comp_01:	ld	tempreg1,Y+		;get char
		cpi	tempreg1,0		;EOL
		brne	lce_comp_02		;no

lce_comp_end:	clr	ereg			;clear error
		ldi	r17,0xff		;write EOL
		st	X+,r17
    		ret

lce_comp_02:	cpi	r18,0x01		;comment mode
		brne	lce_comp_03		;no				
		
lce_comp_wc:	st	X+,tempreg1		;write char
		mov	r19,tempreg1
		inc	r17			;increment destcounter
		rjmp	lce_comp_loop		;next char

lce_comp_03:	cpi	r18,0x02		;string mode
		brne	lce_comp_04
		cpi	tempreg1,'"'		;end of string
		brne	lce_comp_wc		;goto loop
		clr	r18			;switch to normal mode
		rjmp	lce_comp_wc		;goto loop
	
lce_comp_04:	cpi	tempreg1,'"'		;start of string
		brne	lce_comp_05		;no
		ldi	r18,0x02		;switch to string mode
		rjmp	lce_comp_wc		;goto loop
	
lce_comp_05:	cpi	tempreg1,0x20		;space
		brne	lce_comp_05a    
		cpi	r19,0x80
		brcc	lce_comp_loop		;ignore spaces 
		rjmp	lce_comp_wc		;goto loop

lce_comp_05a:	cpi	tempreg1,0x27		;start of comment
		brne	lce_comp_06		;no
		ldi	r18,0x01		;switch to comment mode
		rjmp	lce_comp_wc		;goto loop

		
lce_comp_06:	cpi	tempreg1,'A'		;A-char
		brcs	lce_comp_wc		;less
		cpi	tempreg1,'Z'+1		;Z-char
		brcc	lce_comp_wc		;greater
		ld	tempreg2,Y		;get next char
	    	cpi	tempreg2,'A'		;A-char
		brcs	lce_comp_wc		;less
		cpi	tempreg2,'Z'+1		;Z-char
		brcc	lce_comp_wc		;greater
		
		cpi	r19,0x20
		breq	lce_comp_08
		cpi	r19,':'
		brne	lce_comp_wc	

;we search for token
lce_comp_08:	ldi	tempreg4,0xc0		;first (abbreviated) token
		ldi	ZL,LOW(lce_toktab*2)	;Start of keyword table
		ldi	ZH,HIGH(lce_toktab*2)

lce_comp_10:	lpm	tempreg3,Z+		;get rom char
		cpi	tempreg3,0		;end?
		breq	lce_comp_wc		;yes-> write char
		rjmp	lce_comp_12
		
lce_comp_wsp:	lpm	tempreg3,Z+
		cpi	tempreg3,32		;wait for space in rom table
		brne	lce_comp_wsp
		inc	tempreg4		;token +1
		rjmp	lce_comp_10		;next test
		
lce_comp_12:	cp	tempreg3,tempreg1	;1st char OK?
		brne	lce_comp_wsp		;no->loop
		lpm	tempreg3,Z+
		cp	tempreg3,tempreg2	;2nd char OK?
		brne	lce_comp_wsp		;no
		
;token found
		adiw	YL,1			;point to next char
lce_comp_20:	lpm	tempreg2,Z+		;next ROM char
		cpi	tempreg2,0x20		;end of ROM kw
		brne	lce_comp_21		;no
		andi	tempreg4,0xbf		;clear abbrev bit
lce_comp_22:	mov	tempreg1,tempreg4	;copy
		rjmp	lce_comp_wc		;write char

lce_comp_21:	ld	tempreg1,Y+		;next RAM char
		cpi	tempreg1,'A'		;A-char
		brcs	lce_comp_22		;less
		cpi	tempreg1,'Z'+1		;Z-char
		brcc	lce_comp_22		;greater
		
		cp	tempreg1,tempreg2	;compare
		breq	lce_comp_20		;next char
		
		sbiw	YL,2
		ld	tempreg1,Y+
		rjmp	lce_comp_wc		;write char
		
;-----------------------------------------------------------------------
; expand linebuf -> inbuf
;-----------------------------------------------------------------------
lce_xpand:	ldi	XL,LOW(bas_inbuf+3)	;text buffer address
		ldi	XH,HIGH(bas_inbuf+3)
		ldi	YL,LOW(bas_linebuf)	;code buffer address
		ldi	YH,HIGH(bas_linebuf)
		rjmp	lce_xpand_000
lce_xpand_00:	st	X+,tempreg4		;write char
lce_xpand_000:	ld	tempreg4,Y+		;get char
		cpi	tempreg4,0xff		;EOL
		brne	lce_xpand_01		;no
		st	Y,byte0			;write EOL
		ret				;end
		
lce_xpand_01:	sbrs	tempreg4,7		;skip if token
		rjmp	lce_xpand_00		;char
		
		ldi	ZL,LOW(lce_toktab*2)	;Start of keyword table
		ldi	ZH,HIGH(lce_toktab*2)
		mov	tempreg3,tempreg4	;save token
		andi	tempreg4,0x3f		;64 token

lce_xpand_02:	cpi	tempreg4,0x00
		breq	lce_xpand_04
lce_xpand_03:	lpm	tempreg2,Z+		;get char
		cpi	tempreg2,0x20		;space		
		brne	lce_xpand_03
		dec	tempreg4
		brne	lce_xpand_02

lce_xpand_04:	lpm	tempreg2,Z+		;get ROM char 1
		st	X+,tempreg2		;write	
		lpm	tempreg2,Z+		;get ROM char 2
		st	X+,tempreg2		;write	
		ldi	tempreg2,0x20		;space
		sbrs	tempreg3,6		;abbrev?
lce_xpand_05:	lpm	tempreg2,Z+
		st	X+,tempreg2
		cpi	tempreg2,0x20		;end of keyword
		brne	lce_xpand_05
		rjmp	lce_xpand_000		

		
;------------------------------------------------------------------------------
; write line [ctrl] from inbuf to vram
;------------------------------------------------------------------------------
lce_wline:	rcall	lce_calc		;set address in vram
		ldi	YL,LOW(bas_inbuf)	;text buffer address
		ldi	YH,HIGH(bas_inbuf)
		ldi	tempreg2,libmio_cols	;number of chars to copy
lce_wline1:	ld	tempreg1,Y+		;get from vram
		st	X+,tempreg1
		dec	tempreg2
		brne	lce_wline1		
		ret

;------------------------------------------------------------------------------
; read line [ctrl] from vram to inbuf to vram
;------------------------------------------------------------------------------
lce_rline:	rcall	lce_calc		;set address in vram
		ldi	YL,LOW(bas_inbuf)	;text buffer address
		ldi	YH,HIGH(bas_inbuf)
		ldi	tempreg2,libmio_cols	;number of chars to copy
lce_rline1:	ld	tempreg1,X+		;get from vram
		st	Y+,tempreg1
		dec	tempreg2
		brne	lce_rline1		
		st	Y+,byte0		;add an extra EOL
		ret

;------------------------------------------------------------------------------
; clear inbuf and write line number of [ctrl]
;------------------------------------------------------------------------------
lce_clear:	ldi	ZL,LOW(bas_inbuf)	;text buffer address
		ldi	ZH,HIGH(bas_inbuf)
		ldi	tempreg4,0x20		;space
		ldi	tempreg3,libmio_cols
lce_clear1:	st	Z+,tempreg4
		dec	tempreg3
		brne	lce_clear1	
		
lce_number:	ldi	ZL,LOW(bas_inbuf)	;text buffer address
		ldi	ZH,HIGH(bas_inbuf)
		ldi	tempreg4,0x30		;"0"
		mov	tempreg3,ctrl
		inc	tempreg3		;line numbers start with 1
lce_number1:	subi	tempreg3,10		;-10
		brcs	lce_number2
		inc	tempreg4
		rjmp	lce_number1
lce_number2:	subi	tempreg3,0xc6		;+70
		st	Z+,tempreg4		;10^1
		st	Z+,tempreg3		;10^0
		ldi	tempreg4,0x20		;space
		st	Z+,tempreg4		;write
		ret

;------------------------------------------------------------------------------
; calc line adr of ctrl to X
;------------------------------------------------------------------------------
lce_calc:	ldi	XH,libmio_cols		;bytes/vram-line
		mul	XH,ctrl
	    	ldi	XL,LOW(libmio_vram+3*libmio_cols)	;lines start
		ldi	XH,HIGH(libmio_vram+3*libmio_cols)
		add	XL,r0			;calc line address
		adc	XH,r1
		ret

;------------------------------------------------------------------------------
; shrink inbuf
;------------------------------------------------------------------------------
lce_shrink:	ldi	YL,LOW(bas_inbuf+libmio_cols)
		ldi	YH,HIGH(bas_inbuf+libmio_cols)
		ldi	tempreg2,libmio_cols-3	;max chars
lce_shrink1:	ld	tempreg1,-Y		;		
		cpi	tempreg1,0x20		;space
		brne	lce_shrink2
		st	Y,byte0			;write EOL	
		dec	tempreg2
		brne	lce_shrink1
lce_shrink2:	ret		
			
;------------------------------------------------------------------------------
; load program from EEPROM
;------------------------------------------------------------------------------
lce_pload:	clr	ctrl			;first line is 0
lce_pload1:	call	ee_readline0		;get from EEPROM
		rcall	lce_clear		;clear inbuf
		rcall	lce_xpand		;expand line
		rcall	lce_wline		;write into vram
		inc	ctrl			;next line
		cpi	ctrl,bas_lines		;all lines done
		brne	lce_pload1		;no
		ret

;------------------------------------------------------------------------------
; save program to EEPROM
;------------------------------------------------------------------------------
lce_psave:	clr	ctrl			;first line is 0
lce_psave1:	rcall	lce_rline		;get line from vram
		rcall	lce_shrink		;set EOL
		rcall	lce_comp		;compress line
		call	ee_writeline		;write to eeprom
		inc	ctrl			;next line
		cpi	ctrl,bas_lines		;all lines done
		brne	lce_psave1		;no
		ret

;------------------------------------------------------------------------------
; list program
;------------------------------------------------------------------------------
lce_plist:	clr	ctrl			;first line is 0
lce_plist1:	call	ee_readline0		;get from EEPROM
		rcall	lce_clear		;clear inbuf
		rcall	lce_xpand		;expand line
		rcall	lce_shrink		;set EOL
		push	ctrl
		ldi	ZL,LOW(bas_inbuf)	;set pointer
		ldi	ZH,HIGH(bas_inbuf)
		libmio_ramtext			;output text
		libmio_newline
		pop	ctrl
		inc	ctrl			;next line
		cpi	ctrl,bas_lines		;all lines done
		brne	lce_plist1		;no
		ret


;------------------------------------------------------------------------------
; clear editor lines (new program)
;------------------------------------------------------------------------------
lce_pnew:	clr	ctrl			;first line is 0
lce_pnew1:	rcall	lce_clear		;clear inbuf
		rcall	lce_wline		;write into vram
		inc	ctrl
		cpi	ctrl,bas_lines		;all lines done
		brne	lce_pnew1		;no
		ret

;------------------------------------------------------------------------------
; renumber lines
;------------------------------------------------------------------------------
lce_renumber:	clr	ctrl			;first line is 0
lce_renumber1:	rcall	lce_rline		;read from vram
		rcall	lce_number		;set correct number
		rcall	lce_wline		;write into vram
		inc	ctrl
		cpi	ctrl,bas_lines		;all lines done
		brne	lce_renumber1		;no
		ret

;------------------------------------------------------------------------------
; delete line [ctrl] in vram
;------------------------------------------------------------------------------
lce_del:	cpi	ctrl,bas_lines-1	;last line
		brcs	lce_del0
		ret
lce_del0:	rcall	lce_calc		;get line addr to X
		mov	YL,XL
		mov	YH,XH
lce_del1:	ldi	tempreg2,libmio_cols	;chars per line
lce_del2:	ldd	tempreg1,Y+libmio_cols	;char of next line
		st	Y+,tempreg1		;write to recent line
		dec	tempreg2
		brne	lce_del2		;char loop
		inc	ctrl			;next line
		cpi	ctrl,bas_lines-1
		brne	lce_del1		;line loop
		rcall	lce_clear		;clear buffer line
		rcall	lce_wline
		rjmp	lce_renumber		;renumber lines 
		
;------------------------------------------------------------------------------
; insert line [ctrl] in vram
;------------------------------------------------------------------------------
lce_ins:	cpi	ctrl,bas_lines-1	;last line
		brcs	lce_ins0
		ret
lce_ins0:	mov	tempreg3,ctrl		;copy 
		ldi	ctrl,bas_lines-1		
		rcall	lce_calc		;get line of line max-1
		mov	YL,XL
		mov	YH,XH
lce_ins1:	ldi	tempreg2,libmio_cols	;chars per line
lce_ins2:	ld	tempreg1,-Y		;char of this line
		std	Y+libmio_cols,tempreg1
		dec	tempreg2
		brne	lce_ins2		;char loop
		dec	ctrl			;next line
		cp	ctrl,tempreg3
		brne	lce_ins1		;line loop
		rcall	lce_calc		;get line of line max-1
		rcall	lce_clear		;clear last line
		rcall	lce_wline
		rjmp	lce_renumber		;renumber lines 

;------------------------------------------------------------------------------
; token table
;------------------------------------------------------------------------------
lce_toktab:	.db "PRINT CLS GOTO BOX FBOX SYNC END IF THEN POS NOTE EMIT"
		.db " WKEY RKEY CALL RETURN FOR NEXT INPUT PLOT WAIT TSET"
		.db " TGET EPUT EGET SPUT SGET WPOKE DIR OUT COLOR LCHAR "
		.db "LPIX PUMP BPOKE YEMIT LIMIT GCHAR DRAW DFS SHOW HIDE"
		.db " XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX ",0
								